#ifndef CONSTRUCT_HPP
#define CONSTRUCT_HPP


/** Module for LCP Interval construction 


*/


#include "sdsl/sdsl_concepts.hpp"
#include <sdsl/suffix_trees.hpp>
#include "sdsl/int_vector.hpp"
#include "sdsl/bit_vectors.hpp"
#include <sdsl/construct.hpp>
#include "sdsl/construct_lcp.hpp"
#include "common.hpp"   // Save lcp construction
#include <sdsl/construct_bwt.hpp>
#include "sdsl/construct_sa.hpp"
#include "sdsl/suffix_arrays.hpp"
#include <string>
#include <vector>
#include <tuple>
#include <chrono>
#include <queue>
#include <algorithm>    // std::sort

#define MINOCC 2


typedef sdsl::bit_vector::size_type size_type;
typedef std::tuple<uint64_t,uint64_t> tii;
typedef std::vector<tii> vtii;

struct element
{
	uint64_t begin;
	uint64_t lcp;
	element(uint64_t _begin, uint64_t _lcp)
	{
		begin = _begin;
		lcp   = _lcp;
	}
};

struct lcpinterval
{
	uint32_t l;
	uint32_t lb;
	uint32_t rb;
	bool leftmax;
	
	bool use;
	uint32_t ld;
	
	lcpinterval(uint64_t _l=0, uint64_t _lb=0, uint64_t _rb=0, bool _leftmax=true): 
	    l(_l), lb(_lb), rb(_rb), leftmax(_leftmax)
	{
	}
	friend std::ostream& operator<<(std::ostream& os, const lcpinterval& li)
	{
		os << std::setw(2) << li.l << "-[" << std::setw(2) << li.lb << "," << std::setw(2) << li.rb << "]";
		return os;
	}
};

bool operator>(const lcpinterval& lhs, const lcpinterval& rhs)
{
	return lhs.l < rhs.l;             // Len only 
    //return (lhs.l-1)*(lhs.rb-lhs.lb) < (rhs.l-1)*(rhs.rb-rhs.lb);   // Area
}

typedef std::priority_queue<lcpinterval, std::vector<lcpinterval>, std::greater<lcpinterval>> pqli;  // longest first
//typedef queue<lcpinterval> pqli;                                                      // Bottom up
typedef std::vector<lcpinterval> vli;

/** Check wether the given lcp interval is allowed  */
bool allowed(const uint64_t sa_i, const uint64_t sa_j, const lcpinterval& li, const lcpinterval& prev, uint64_t& k, uint64_t& t)
{
	// check for overlap repeat:
	uint64_t b = li.rb-li.lb+1;
	
	if (b == 2) // Check for new overlap
	{
	    if (sa_i + li.l > sa_j)
	    {
	        k = 1;
	        t = sa_j - sa_i;	        
	    }
	    if (sa_j + li.l > sa_i)
	    {
	        k = 1;
	        t = sa_i - sa_j;
	    }
	    return true;
	}
	
	if (li.l + t == prev.l) // Prev overlap
	{
	    uint64_t prev_b = prev.rb-prev.lb+1;
	    if (prev_b + k != b)    // New occ
	    {
	        k = b - prev_b;
	        return true;
	    }
	    return false; 	     // prev_b + k == b
	}    
    return true;
}


/** Traverse the LCP array bottom up and save all allowed lcp intervals in a priority queue */
pqli bu_lcp(sdsl::cache_config& config, uint64_t minlen)
{
	sdsl::int_vector_buffer<> lcp(sdsl::cache_file_name(sdsl::conf::KEY_LCP, config));
	sdsl::int_vector_buffer<> sa(sdsl::cache_file_name(sdsl::conf::KEY_SA, config));
	sdsl::int_vector_buffer<8> bwt(sdsl::cache_file_name(sdsl::conf::KEY_BWT, config));


	pqli pq;

	std::stack<element> st;
	st.push(element(0, 0));
	
	lcpinterval prev;
    
    uint64_t size = 0;
 
    uint64_t lcp_size = lcp.size();
    uint64_t prev_sa = sa[0];
    uint64_t sa_i = sa[0];
    
    uint64_t k = 0;
    uint64_t t = 0;

    uint64_t last = 0;
    char last_bwt = bwt[0];
    uint64_t count = 0;
        
	for(uint64_t i=1; i<lcp_size; ++i)
	{
		uint64_t lb = i-1;
		uint64_t lcp_value = lcp[i];
		sa_i = sa[i];
		while(st.top().lcp > lcp_value)
		{
			lcpinterval li(st.top().lcp, st.top().begin, i-1);
			st.pop();
				
		    if (last <= li.lb)
		    {
		        li.leftmax = false;		
		        count++;	    
		    }
						
			if (li.l > minlen and allowed(sa_i, prev_sa, li, prev, k, t))
			{ 
			    if (li.leftmax)
			    {
        			pq.push(li);
        			size++;
    			}
       		}
			prev.l = li.l; prev.lb = li.lb; prev.rb = li.rb;
			
			lb = li.lb;
		}
		if(st.top().lcp < lcp_value)
		{
			st.push(element(lb, lcp_value));
		}
		
		prev_sa = sa_i;
		
		char bwt_i = bwt[i];
		if (bwt_i != last_bwt)
		{
		    last = i;
		    last_bwt = bwt_i;	
		}	
	}
	while(!st.empty())
	{
			lcpinterval li(st.top().lcp, st.top().begin, lcp.size()-1);
			st.pop();
			
				
		    if (last <= li.lb)
		    {
		        li.leftmax = false;		
		        count++;	    
		    }
						
			if (li.l > minlen and allowed(sa_i, prev_sa, li, prev, k, t))
			{
			    if (li.leftmax)
			    {
        			pq.push(li);
        			size++;
    			}
    		}
			prev.l = li.l; prev.lb = li.lb; prev.rb = li.rb;
	}
	return pq;
}

typedef std::tuple<uint64_t, uint64_t, uint64_t> tiii;
bool tuplesort(tiii a, tiii b)
{
    return std::get<1>(a) < std::get<1>(b);
}

/** Construct the needed structures (i.e. Suffix Array, BWT and LCP array) and safe with sdsl library (file cache) */
sdsl::cache_config construct(std::string file)
{
	// Get parameters
	uint8_t num_bytes = 1; // Byte Alphabet
	std::string dir = ".";
	std::string id = "out";
	sdsl::cache_config config(false, dir, id);

	{// (1) check, if the text is cached
		if ( !sdsl::cache_file_exists(sdsl::conf::KEY_TEXT, config) ){
		    sdsl::int_vector<8> text;
			sdsl::load_vector_from_file(text, file, num_bytes);
			if (false)
			{
				append_zero_symbol(text);
			}
			sdsl::store_to_cache(text, sdsl::conf::KEY_TEXT, config);
		}
		sdsl::register_cache_file(sdsl::conf::KEY_TEXT, config);
	}

	{// (2) check, if the suffix array is cached
		if ( !sdsl::cache_file_exists(sdsl::conf::KEY_SA, config) )
		{
			sdsl::construct_sa<8>(config);
		}
		sdsl::register_cache_file(sdsl::conf::KEY_SA, config);
	}

	{//  (3) construct BWT
		if ( !sdsl::cache_file_exists(sdsl::conf::KEY_BWT, config) )
		{
			sdsl::construct_bwt<8>(config);
		}
		sdsl::register_cache_file(sdsl::conf::KEY_BWT, config);
	}

	{//  (4) construct LCP
		if ( !sdsl::cache_file_exists(sdsl::conf::KEY_LCP, config) )
		{
			construct_lcp_semi_extern_PHI_save(config);
		}
		sdsl::register_cache_file(sdsl::conf::KEY_LCP, config);
	}
	return config;
}


#endif
