#include <string>
#include <iostream>
#include <sstream>

#include "construct.hpp"
#include "compress_basic.hpp"
#include "decompress_basic.hpp"
#include "compress_advanced.hpp"
#include "decompress_advanced.hpp"

using namespace std;
using namespace sdsl;

inline bool exists (const std::string& name) {
    if (FILE *file = fopen(name.c_str(), "r")) {
        fclose(file);
        return true;
    } else {
        return false;
    }   
}

struct Parameters
{
    int a, b;       // Parameters as in the paper. An interval will be checked if l >= a/size + b, where size is the size of the factor
    string comp;    // The final compression algorithm (one of "xz-9", "zpaq", "ppmd"). The choosen programm has to be installed
    string input;   // The name of the input file
    string output;  // The name of the output file. The appropriate ending (see Parameter comp) will be added
    bool advanced;  // Wether to take the advanced version
    bool compress;  // Compress or decompress
};

void usage(int argc, char** argv)
{
    cerr << "Usage: " << argv[0] << " [options] <Inputfile> [Outputfile]" << endl << endl;
    cerr << "If Outputfile is not given, it will use the name of the Inputfile adding/removing the fileextension '.tar' (if compressing / decompressing)" << endl;
    cerr << "Valid options are: " << endl;
    cerr << "-h          => Show this help" << endl;
    cerr << "-a [value]  => The value for the parameter a (see below), optional, default = 30" << endl;
    cerr << "-b [value]  => The value for the paramteer b (see below), optional, default = 80" << endl;
    cerr << "-d          => If given, the algorithm will decompress, optional" << endl;
    cerr << "-e          => If given, the algorithm will use the advanced version, optional" << endl;
    cerr << "-c [value]  => The final compression method, one of {xz-9, zpaq, ppmd}, optional, default = xz-9" << endl << endl;
    cerr << "The parameters a and b are further discussed in the paper. An interval will be checked, if l >= a/size + b, where size is the size of the interval" << endl;
    exit(1);
}

bool has_suffix(const std::string &str, const std::string &suffix)
{
    return str.size() >= suffix.size() &&
           str.compare(str.size() - suffix.size(), suffix.size(), suffix) == 0;
}

Parameters parse(int argc, char** argv)
{
    Parameters p;
    p.a = 30; p.b = 80;
    p.comp = "xz-9";
    p.advanced = false; 
    p.compress = true;
    
    int i = 1;
    bool param = true;
    while (i < argc)
    {
        char* next = argv[i];
        param = next[0] == '-';
        if (not param)
            break;
        
        int j = 1;
        while (next[j])
        {
            if (next[j] == 'h')
                usage(argc, argv);
               
            if (next[j] == 'd')
                p.compress = false;
            
            if (next[j] == 'e')
                p.advanced = true;
                
            if (next[j] == 'a')
            {
                p.a = std::atoi(argv[i+1]);
                i++;
                break;
            }  
            if (next[j] == 'b')
            {
                p.b = std::atoi(argv[i+1]);
                i++;
                break;
            }  
                
            if (next[j] == 'c')
            {
                p.comp = std::string(argv[i+1]);
                i++;
                break;
            }  
                
            j++;
        }
        i++;
    }
    
    if (i >= argc)
    {
        std::cerr << "Missing Inputfile." << std::endl;
        usage(argc, argv);
    }
    
    p.input = std::string(argv[i++]);
    
    if (not exists(p.input))
    {
        std::cerr << "Cannot access file " << p.input << "." << std::endl;
        usage(argc, argv);
    }
    
    std::string filesuff;
    if (p.comp == "xz-9")
        filesuff = "xz";
    else if (p.comp == "zpaq")
        filesuff = "zpaq";
    else if (p.comp == "ppmd")
        filesuff = "ppmd";
    else
    {
        std::cerr << "Unknown compression method: " << p.comp << std::endl;
        usage(argc, argv);
    }
    
    if (i >= argc)
    {
        if (p.compress)
            p.output = p.input + ".tar";
        
        else
        {
            size_t lastindex = p.input.find_last_of("."); 
            p.output = p.input.substr(0, lastindex); 
        }
    }
    else
    {
        p.output = std::string(argv[i]);
        //if (not has_suffix(p.output, filesuff))
        //    p.output += filesuff;
    }
    return p;
}


int main(int argc, char **argv)
{	
	Parameters p = parse(argc, argv);
	
	std::cout << "======================================" << std::endl;
	std::cout << "Parameters are: " << std::endl;
	std::cout << "a: " << p.a << std::endl;
	std::cout << "b: " << p.b << std::endl;
	std::cout << "compressionmethod: " << p.comp << std::endl;
	std::cout << "input: " << p.input << std::endl;
	std::cout << "output: " << p.output << std::endl;
	std::cout << "Compress: " << p.compress << std::endl;
	std::cout << "Advanced: " << p.advanced << std::endl;
	std::cout << "======================================" << std::endl;
	
	
	
	if (p.compress)
	{
        cache_config config = construct(p.input);
	    
	    if (p.advanced)
	    {
	        compress_advanced(config, p.output, p.a, p.b, p.comp);
	    }   
        else
        {
            compress_basic(config, p.output, p.a, p.b, p.comp);
        }
        
        {
		    remove(cache_file_name(conf::KEY_TEXT, config));
		    remove(cache_file_name(conf::KEY_SA, config));
		    remove(cache_file_name(conf::KEY_ISA, config));
		    remove(cache_file_name(conf::KEY_BWT, config));
		    remove(cache_file_name(conf::KEY_LCP, config));
        }
	}
	else
	{
	    if (p.advanced)
	    {   
	        decompress_advanced(p.input, p.output, p.comp);
	    }
	    else
	    {
	        decompress_basic(p.input, p.output, p.comp);
	    }
	}
    
    return 0;

}

